package cn.ecpark.service.usergw.config;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.BiConsumer;

import com.ctrip.framework.apollo.internals.ConfigRepository;
import com.ctrip.framework.apollo.internals.DefaultConfig;
import com.ctrip.framework.apollo.internals.PropertiesCompatibleFileConfigRepository;
import com.ctrip.framework.apollo.internals.YmlConfigFile;

import org.apache.dubbo.config.ReferenceConfig;
import org.apache.dubbo.rpc.service.GenericService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cloud.gateway.filter.FilterDefinition;
import org.springframework.cloud.gateway.handler.predicate.PredicateDefinition;
import org.springframework.cloud.gateway.route.RouteDefinition;
import org.springframework.cloud.gateway.route.RouteDefinitionLocator;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.stereotype.Component;
import org.yaml.snakeyaml.Yaml;

import cn.ecpark.service.usergw.biz.filters.bean.GenericServicePool;
import cn.ecpark.service.usergw.utils.Extract;
import reactor.core.publisher.Flux;

/**
 * ConfigGateway
 */
@Component
public class ConfigGateway implements RouteDefinitionLocator {

  // List<MediaType> mediaTypes = new ArrayList<>();
  private static final org.slf4j.Logger log = org.slf4j.LoggerFactory.getLogger(ConfigGateway.class);

  @Value("${yame.config:}")
  private String yameConfigPath;



  private List<FilterDefinition> defaultFilters = new ArrayList<>();
  private HashMap<String, BiConsumer<ReferenceConfig<GenericService>, Object>> specialField = new HashMap<String, BiConsumer<ReferenceConfig<GenericService>, Object>>();
  private Set<String> ignoreKey = new HashSet<>();

  /**
   * spring boot 上下文
   */
  @Autowired
  private ApplicationContext appContext;

  /**
   * ConfigGateway 构造函数
   */
  public ConfigGateway() {
    specialField.put("application", ConfigSpecialFunction::setApplication);
    specialField.put("registry", ConfigSpecialFunction::setRegistry);
    specialField.put("methods", ConfigSpecialFunction::setMethods);

    ignoreKey.add("predicates");
    ignoreKey.add("filters");
  }

  @Override
  @SuppressWarnings("unchecked")
  public Flux<RouteDefinition> getRouteDefinitions() {
    defaultFilters.clear();
    InputStream inputStream;

    ConfigApollo configApollo = appContext.getBean(ConfigApollo.class);
    configApollo.Connect();
    if(configApollo.isConnected()) {
      inputStream = loadApolloConfig(configApollo);
    } else {
      inputStream = loadLocalFileConfig();
    }
    
    if (inputStream != null) {
      Map<String, Object> configYaml = new Yaml().load(inputStream);
      if (configYaml != null) {

        List<String> defaultFiltersYaml = null;
        Map<String, Object> restfulYaml = null;
        Map<String, Object> dubboYaml = null;

        Object unknown;
        unknown = configYaml.get("default-filters");
        if (unknown != null) {
          defaultFiltersYaml = (List<String>) unknown;
        }

        unknown = configYaml.get("restful");
        if (unknown != null) {
          restfulYaml = (Map<String, Object>) unknown;
        }

        unknown = configYaml.get("dubbo");
        if (unknown != null) {
          dubboYaml = (Map<String, Object>) unknown;
        }

        List<RouteDefinition> routeList = new ArrayList<RouteDefinition>();
        if (restfulYaml != null && !restfulYaml.isEmpty()) {
          this.getDefaultFilter(defaultFilters, defaultFiltersYaml);
          this.configRestful(routeList, restfulYaml);
        }

        if (dubboYaml != null && !dubboYaml.isEmpty()) {
          this.configHttp2Dubbo(routeList, dubboYaml);
        }

        if (!routeList.isEmpty()) {
          return Flux.fromIterable(routeList);
        }
      }
    }

    return Flux.empty();
  }

  private InputStream loadLocalFileConfig() {
    Object inputStream = null;
    String[] gatewayConfigPathList;
    if (yameConfigPath.equals("")) {
      gatewayConfigPathList = new String[] { "gateway.yml", "gateway.yaml", "Gateway.yml", "Gateway.yaml" };
    } else {
      gatewayConfigPathList = new String[] { yameConfigPath };
    }
    for (String gatewayConfigPath : gatewayConfigPathList) {
      gatewayConfigPath = gatewayConfigPath.trim();
      inputStream = this.getClass().getClassLoader().getResourceAsStream(gatewayConfigPath);
      if (inputStream != null) {
        break;
      }
    }
    return (InputStream) inputStream;
  }
  
  
  private InputStream loadApolloConfig(ConfigApollo configApollo) {
    DefaultConfig config = (DefaultConfig) configApollo.getConfig();
    try {
      Field m_configRepository = DefaultConfig.class.getDeclaredField("m_configRepository");
      m_configRepository.setAccessible(true);
      ConfigRepository cr = (ConfigRepository) m_configRepository.get(config);
      Field configFile = PropertiesCompatibleFileConfigRepository.class.getDeclaredField("configFile");
      configFile.setAccessible(true);
      YmlConfigFile ymlConfigFile = (YmlConfigFile) configFile.get(cr);
      String content = ymlConfigFile.getContent();
      log.info("Load Apollo YAML Config: \n{}", content);
      return new ByteArrayInputStream(content.getBytes());
    } catch (NoSuchFieldException e) {
      e.printStackTrace();
    } catch (SecurityException e) {
      e.printStackTrace();
    } catch (IllegalArgumentException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * 
   * @param routeList   路由列表
   * @param restfulYaml 配置Yaml.restful对象
   */
  @SuppressWarnings("unchecked")
  private void configRestful(List<RouteDefinition> routeList, Map<String, Object> restfulYaml) {

    Object unknownRoutes = restfulYaml.get("routes");
    if (unknownRoutes != null) {
      List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
      for (LinkedHashMap<String, List<String>> iter : routes) {

        List<FilterDefinition> filters = new ArrayList<>();
        List<PredicateDefinition> predicates = new ArrayList<>();
        RouteDefinition rd = new RouteDefinition();

        // 设置基础属性
        this.parseAndSetBase(rd, iter);

        // predicates: 下的相关属性
        this.parseAndAddPredicates(predicates, iter, "predicates");

        // filters: 下的相关属性
        this.parseAndAddFilters(filters, iter, "filters");

        rd.setPredicates(predicates);
        rd.setFilters(filters);

        routeList.add(rd);
      }
    }

  }

  /**
   * 
   * @param routeList   路由列表
   * @param configDubbo 配置Yaml.dubbo对象
   */
  @SuppressWarnings("unchecked")
  private void configHttp2Dubbo(List<RouteDefinition> routeList, Map<String, Object> configDubbo) {

    try {

      Object unknownRoutes = configDubbo.get("routes");
      if (unknownRoutes != null) {

        List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
        for (LinkedHashMap<String, List<String>> iter : routes) {

          List<FilterDefinition> filters = new ArrayList<>();
          List<PredicateDefinition> predicates = new ArrayList<>();
          RouteDefinition rd = new RouteDefinition();

          // 设置基础属性
          String dubboUri = this.parseDubboUriAndSetBase(rd, iter);

          // predicates: 下的相关属性
          this.parseAndAddPredicates(predicates, iter, "predicates");

          // filters: 下的相关属性
          this.parseAndAddDubboFilters(dubboUri, filters, iter, "filters");

          rd.setPredicates(predicates);
          rd.setFilters(filters);

          routeList.add(rd);

        }

      }

    } catch (Exception e) {
      log.error(e.getMessage());
      log.warn("App is exit");
      ((ConfigurableApplicationContext) appContext).close();
    }

    // gs.$invoke(method, parameterTypes, args)

  }

  /**
   *
   * @param filters     {@link #defaultFilters} 全局的默认过滤列表, 可以和Spring Cloud
   *                    默认配置一起生效
   * @param defaultYaml 配置Yaml.default-filters对象
   */
  private void getDefaultFilter(List<FilterDefinition> filters, List<String> defaultFiltersYaml) {
    // default-filters: 下的相关设置
    if (defaultFiltersYaml != null) {
      for (String filterString : defaultFiltersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }
  }

  /***
   * 
   * @param rd   路由列表
   * @param iter Yaml.restful.routes
   */
  private void parseAndSetBase(RouteDefinition rd, LinkedHashMap<String, List<String>> iter) {
    // 设置id
    Object id = iter.get("id");
    if (id != null) {
      rd.setId((String) id);
    }

    // 设置uri
    Object uri = iter.get("uri");
    if (uri != null) {
      rd.setUri(URI.create((String) uri));
    }

    // 设置uri
    Object order = iter.get("order");
    if (order != null) {
      rd.setOrder((int) order);
    } else {
      rd.setOrder(0);
    }
  }

  /**
   * 
   * @param rd   {@link RouteDefinition}
   * @param iter Yaml.dubbo.routes
   * @return dubboUri 字符串标识 eg. dubbo://application/group/com.a.b:1.0.0
   */
  @SuppressWarnings("unchecked")
  private String parseDubboUriAndSetBase(RouteDefinition rd, LinkedHashMap<String, List<String>> iter) {
    // 设置id
    Object id = iter.get("id");
    if (id != null) {
      rd.setId((String) id);
    }
    iter.remove("id");

    // 设置uri
    Object uri = iter.get("uri");
    if (uri != null) {
      try {
        rd.setUri(URI.create((String) uri));
      } catch (Exception e) {
        log.warn("dubbo Uri = {} error, so set uri to default value dubbo://yame", uri);
        rd.setUri(URI.create("dubbo://yame"));
      }
    } else {
      rd.setUri(URI.create("dubbo://yame"));
    }
    iter.remove("uri");

    // 设置uri
    Object order = iter.get("order");
    if (order != null) {
      rd.setOrder((int) order);
      iter.remove("order");
    } else {
      rd.setOrder(0);
    }

    ReferenceConfig<GenericService> reference = new ReferenceConfig<GenericService>();
    reference.setConnections(3);

    String UriString = "dubbo://";

    // TODO: 以后可以完善dubbo-flag形式的判断
    Object methods = iter.get("methods");
    if (methods == null || ((List<Object>) methods).size() == 0) {
      UriString = "dubbo-allowed://";
    } else {
      UriString = "dubbo://";
    }

    Object application = iter.get("application");
    if (application != null) {
      BiConsumer<ReferenceConfig<GenericService>, Object> doFunc = specialField.get("application");
      if (doFunc != null) {
        doFunc.accept(reference, (String) application);
      }
    }
    iter.remove("application");

    // 遍历所有dubbo配置的key, 如果不存在特殊处理的, 就直接反射到reference 默认方法上
    for (Entry<String, List<String>> entry : iter.entrySet()) {
      // Object group = iter.get("group");
      String key = entry.getKey();
      if (ignoreKey.contains(key)) {
        continue;
      }
      Object value = entry.getValue();

      if (value != null) {
        BiConsumer<ReferenceConfig<GenericService>, Object> doFunc = specialField.get((String) key);
        if (doFunc != null) {
          doFunc.accept(reference, value);
        } else {
          try {
            ConfigSpecialFunction.setDefault(reference, key, value);
          } catch (Exception e) {
            log.error("配置解析错误 字段: {}:{}\n{}", key, value, e.toString());
            log.warn("App is exit");
            ((ConfigurableApplicationContext) appContext).close();
          }
        }
      }
    }

    UriString += Extract.getReferenceConfigKey(reference);
    GenericServicePool gsPool = appContext.getBean(GenericServicePool.class);
    reference.setGeneric(true);
    gsPool.put(UriString, reference.get());

    return UriString;
  }

  /**
   * 
   * @param predicates List<{@link PredicateDefinition}> 断言列表
   * @param iter       restful.routes.predicates
   * @param yamlField  根据域的key 获取 predicates 字符串列
   */
  private void parseAndAddPredicates(List<PredicateDefinition> predicates, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> predicatesYaml = iter.get(yamlField);
    if (predicatesYaml != null) {
      for (String predicateString : predicatesYaml) {
        predicates.add(new PredicateDefinition(predicateString));
      }
    }
  }

  /**
   * 
   * @param filters   List<{@link FilterDefinition}> 过滤器列表
   * @param iter      restful.routes.filters
   * @param yamlField 根据域的key 获取 filters 字符串列
   */
  private void parseAndAddFilters(List<FilterDefinition> filters, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> filtersYaml = iter.get(yamlField);
    if (filtersYaml != null) {
      filters.addAll(defaultFilters);
      for (String filterString : filtersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }
  }

  /**
   * 
   * @param dubboUri  获取返回的字符串
   *                  {@link #parseDubboUriAndSetBase(RouteDefinition, LinkedHashMap)}
   * @param filters   List<{@link FilterDefinition}> 过滤器列表
   * @param iter      dubbo.routes.filters
   * @param yamlField 根据域的key 获取 filters 字符串列
   */
  @SuppressWarnings("unchecked")
  private void parseAndAddDubboFilters(String dubboUri, List<FilterDefinition> filters,
      LinkedHashMap<String, List<String>> iter, String yamlField) {
    List<String> filtersYaml = iter.get(yamlField);

    filters.addAll(defaultFilters);

    if (filtersYaml != null) {
      for (Object filterObject : filtersYaml) {
        // 现阶段只支持两种情况
        if (filterObject.getClass() == String.class) {
          String filterString = (String) filterObject;
          FilterDefinition fd = new FilterDefinition(filterString);
          log.info(fd.getName());
          if (!fd.getName().equals("Dubbo")) {
            filters.add(fd);
          }
        } else if (filterObject.getClass() == LinkedHashMap.class) {
          Map<String, Object> filterMap = (LinkedHashMap<String, Object>) filterObject;
          FilterDefinition fd = new FilterDefinition();

          fd.setName((String) filterMap.get("name"));
          if (filterMap.containsKey("args")) {
            Object args = filterMap.get("args");
            for (Entry<String, Object> kv : ((Map<String, Object>) args).entrySet()) {
              if (kv.getValue().getClass() != String.class) {
                kv.setValue(String.valueOf(kv.getValue()));
              }
            }
            fd.setArgs((Map<String, String>) args);
          }

          log.info(fd.getName());
          if (!fd.getName().equals("Dubbo")) {
            filters.add(fd);
          }
        }

      }
    }

    filters.add(new FilterDefinition("Dubbo=" + dubboUri));
  }
}